@@ -35,7 +35,9 @@
 #include "fr-command-cfile.h"
 #include "fr-command-lha.h"
 #include "fr-command-rar.h"
+#include "fr-command-rpm.h"
 #include "fr-command-tar.h"
+#include "fr-command-unstuff.h"
 #include "fr-command-zip.h"
 #include "fr-command-zoo.h"
 #include "fr-error.h"
@@ -258,6 +260,12 @@ create_command_from_mime_type (FRArchive  *archive,
 	} else if (is_mime_type (mime_type, "application/x-arj")) {
 		archive->command = fr_command_arj_new (archive->process, 
 						       filename);
+	} else if (is_mime_type (mime_type, "application/x-stuffit")) {
+		archive->command = fr_command_unstuff_new (archive->process,
+							   filename);
+	} else if (is_mime_type (mime_type, "application/x-rpm")) {
+		archive->command = fr_command_rpm_new (archive->process,
+						       filename);
 	} else 
 		return FALSE;
 
@@ -407,6 +415,13 @@ create_command_from_filename (FRArchive  *archive,
 		} else if (file_extension_is (filename, ".lzo")) {
 			archive->command = fr_command_cfile_new (archive->process, filename, FR_COMPRESS_PROGRAM_LZOP);
 			archive->is_compressed_file = TRUE;
+		} else if (file_extension_is (filename, ".bin")
+			   || file_extension_is (filename, ".sit")) {
+			archive->command = fr_command_unstuff_new (archive->process,
+								   filename);
+		} else if (file_extension_is (filename, ".rpm")) {
+			archive->command = fr_command_rpm_new (archive->process,
+							       filename);
 		} else
 			return FALSE;
 	} else
@@ -627,7 +642,10 @@ fr_archive_load (FRArchive   *archive,
 
 	fr_archive_stoppable (archive, TRUE);
 	archive->command->fake_load = fr_archive_fake_load (archive);
+
+	fr_process_clear (archive->process);
 	fr_command_list (archive->command);
+	fr_process_start (archive->process);
 
 	return TRUE;
 }
@@ -641,7 +659,10 @@ fr_archive_reload (FRArchive *archive)
 
 	fr_archive_stoppable (archive, TRUE);
 	archive->command->fake_load = fr_archive_fake_load (archive);
+
+	fr_process_clear (archive->process);
 	fr_command_list (archive->command);
+	fr_process_start (archive->process);
 }
 
 
@@ -671,6 +692,56 @@ fr_archive_rename (FRArchive  *archive,
 /* -- add -- */
 
 
+static char *
+create_tmp_base_dir (const char *base_dir,
+		     const char *dest_path) 
+{
+	char *dest_dir;
+	char *temp_dir;
+	char *tmp;
+	char *parent_dir, *dir;
+
+	if ((dest_path == NULL) 
+	    || (*dest_path == '\0') 
+	    || (strcmp (dest_path, "/") == 0)) 
+		return g_strdup (base_dir);
+
+	dest_dir = g_strdup (dest_path);
+	if (dest_dir[strlen (dest_dir) - 1] == G_DIR_SEPARATOR)
+		dest_dir[strlen (dest_dir) - 1] = 0;
+
+#ifdef DEBUG
+	g_print ("base_dir: %s\n", base_dir);
+	g_print ("dest_dir: %s\n", dest_dir);
+#endif
+
+	temp_dir = get_temp_work_dir_name ();
+	tmp = remove_level_from_path (dest_dir);
+	parent_dir =  g_build_filename (temp_dir, tmp, NULL);
+	g_free (tmp);
+
+	ensure_dir_exists (parent_dir, 0700);
+
+#ifdef DEBUG
+	g_print ("mkdir %s\n", parent_dir);
+#endif
+
+	g_free (parent_dir);
+
+	dir = g_build_filename (temp_dir, "/", dest_dir, NULL);
+	symlink (base_dir, dir);
+
+#ifdef DEBUG
+	g_print ("symlink %s --> %s\n", dir, base_dir);
+#endif
+
+	g_free (dir);
+	g_free (dest_dir);
+
+	return temp_dir;
+}
+
+
 /* Note: all paths unescaped. */
 static FileData *
 find_file_in_archive (FRArchive *archive, 
@@ -748,14 +819,17 @@ void
 fr_archive_add (FRArchive     *archive, 
 		GList         *file_list, 
 		const char    *base_dir,
+		const char    *dest_dir,
 		gboolean       update,
 		const char    *password,
 		FRCompression  compression)
 {
-	GList    *new_file_list;
-	gboolean  free_new_file_list;
+	GList    *new_file_list = NULL;
+	gboolean  base_dir_created = FALSE;
+	gboolean  new_file_list_created = FALSE;
 	GList    *e_file_list;
 	GList    *scan;
+	char     *tmp_base_dir = NULL;
 
 	if (file_list == NULL)
 		return;
@@ -763,23 +837,48 @@ fr_archive_add (FRArchive     *archive,
 	if (archive->read_only)
 		return;
 
+	tmp_base_dir = g_strdup (base_dir);
+
+	if ((dest_dir != NULL) && (*dest_dir != '\0') && (strcmp (dest_dir, "/") != 0)) {
+		const char *rel_dest_dir = dest_dir;
+
+		tmp_base_dir = create_tmp_base_dir (base_dir, dest_dir);
+		base_dir_created = TRUE;
+
+		if (dest_dir[0] == G_DIR_SEPARATOR)
+			rel_dest_dir = dest_dir + 1;
+
+		new_file_list_created = TRUE;
+		new_file_list = NULL;
+		for (scan = file_list; scan != NULL; scan = scan->next) {
+			char *filename = scan->data;
+			new_file_list = g_list_prepend (new_file_list, g_build_filename (rel_dest_dir, filename, NULL));
+		}
+	} else
+		new_file_list = file_list;
+
 	fr_archive_stoppable (archive, FALSE);
 
 	/* if the command cannot update,  get the list of files that are 
 	 * newer than the ones in the archive. */
 
-	free_new_file_list = FALSE;
 	if (update && ! archive->command->propAddCanUpdate) {
-		free_new_file_list = TRUE;
-		new_file_list = newer_files_only (archive, file_list, base_dir);
-	} else
-		new_file_list = file_list;
+		GList *tmp_file_list = new_file_list;
+		new_file_list = newer_files_only (archive, tmp_file_list, tmp_base_dir);
+		if (new_file_list_created)
+			path_list_free (tmp_file_list);
+		new_file_list_created = TRUE;
+	} 
 
 	if (new_file_list == NULL) {
 #ifdef DEBUG
 		g_print ("nothing to update.\n");
 #endif
 
+		if (base_dir_created) 
+			rmdir_recursive (tmp_base_dir);
+		g_free (tmp_base_dir);
+
 		archive->process->error.type = FR_PROC_ERROR_NONE;
 		g_signal_emit_by_name (G_OBJECT (archive->process), 
 				       "done",
@@ -788,8 +887,6 @@ fr_archive_add (FRArchive     *archive,
 		return;
 	}
 
-	fr_process_clear (archive->process);
-
 	fr_command_uncompress (archive->command);
 
 	/* when files are already present in a tar archive and are added
@@ -813,6 +910,7 @@ fr_archive_add (FRArchive     *archive,
 
 		if (del_list != NULL) {
 			_archive_remove (archive, del_list);
+			fr_process_set_ignore_error (archive->process, TRUE);
 			g_list_free (del_list);
 		}
 	}
@@ -841,7 +939,7 @@ fr_archive_add (FRArchive     *archive,
 		prev->next = NULL;
 		fr_command_add (archive->command, 
 				chunk_list, 
-				base_dir, 
+				tmp_base_dir, 
 				update,
 				password,
 				compression);
@@ -849,11 +947,21 @@ fr_archive_add (FRArchive     *archive,
 	}
 
 	path_list_free (e_file_list);
-	if (free_new_file_list)
+	if (new_file_list_created)
 		g_list_free (new_file_list);
 
 	fr_command_recompress (archive->command, compression);
-	fr_process_start (archive->process);
+
+	if (base_dir_created) { /* remove the temp dir */
+		fr_process_begin_command (archive->process, "rm");
+		fr_process_set_working_dir (archive->process, g_get_tmp_dir());
+		fr_process_set_sticky (archive->process, TRUE);
+		fr_process_add_arg (archive->process, "-rf");
+		fr_process_add_arg (archive->process, tmp_base_dir);
+		fr_process_end_command (archive->process);
+
+	}
+	g_free (tmp_base_dir);
 }
 
 
@@ -861,7 +969,7 @@ static void
 file_list_remove_from_pattern (GList      **list, 
 			       const char  *pattern)
 {
-	gchar **patterns;
+	char  **patterns;
 	GList  *scan;
 	
 	if (pattern == NULL)
@@ -882,6 +990,7 @@ file_list_remove_from_pattern (GList      **list,
 			scan = *list;
 		} else
 			scan = scan->next;
+
 		g_free (utf8_name);
 	}
 	
@@ -896,6 +1005,7 @@ typedef struct {
 	FRArchive     *archive;
         char          *exclude_files;
 	char          *base_dir;
+	char          *dest_dir;
 	gboolean       update;
 	char          *password;
 	FRCompression  compression;
@@ -911,17 +1021,18 @@ add_with_wildcard__step2 (GList *file_list, gpointer data)
 
 	file_list_remove_from_pattern (&file_list, aww_data->exclude_files);
 
-	if (aww_data->done_func) 
-		aww_data->done_func (aww_data->done_data);
-
 	fr_archive_add (aww_data->archive,
 			file_list,
 			aww_data->base_dir,
+			aww_data->dest_dir,
 			aww_data->update,
 			aww_data->password,
 			aww_data->compression);
 	path_list_free (file_list);
 
+	if (aww_data->done_func) 
+		aww_data->done_func (aww_data->done_data);
+
 	g_free (aww_data->base_dir);
 	g_free (aww_data->password);
 	g_free (aww_data->exclude_files);
@@ -934,6 +1045,7 @@ fr_archive_add_with_wildcard (FRArchive     *archive,
 			      const char    *include_files,
 			      const char    *exclude_files,
 			      const char    *base_dir,
+			      const char    *dest_dir,
 			      gboolean       update,
 			      gboolean       recursive,
 			      gboolean       follow_links,
@@ -956,6 +1068,7 @@ fr_archive_add_with_wildcard (FRArchive     *archive,
 	aww_data = g_new0 (AddWithWildcardData, 1);
 	aww_data->archive = archive;
 	aww_data->base_dir = g_strdup (base_dir);
+	aww_data->dest_dir = g_strdup (dest_dir);
 	aww_data->update = update;
 	aww_data->password = g_strdup (password);
 	aww_data->compression = compression;
@@ -983,6 +1096,7 @@ typedef struct {
 	FRArchive     *archive;
 	char          *directory;
 	char          *base_dir;
+	char          *dest_dir;
 	gboolean       update;
 	char          *password;
 	FRCompression  compression;
@@ -996,17 +1110,18 @@ add_directory__step2 (GList *file_list, gpointer data)
 {
 	AddDirectoryData *ad_data = data;
 
-	if (ad_data->done_func) 
-		ad_data->done_func (ad_data->done_data);
-
 	fr_archive_add (ad_data->archive,
 			file_list,
 			ad_data->base_dir,
+			ad_data->dest_dir,
 			ad_data->update,
 			ad_data->password,
 			ad_data->compression);
 	path_list_free (file_list);
 
+	if (ad_data->done_func) 
+		ad_data->done_func (ad_data->done_data);
+
 	/**/
 
 	g_free (ad_data->directory);
@@ -1021,6 +1136,7 @@ VisitDirHandle *
 fr_archive_add_directory (FRArchive     *archive, 
 			  const char    *directory,
 			  const char    *base_dir,
+			  const char    *dest_dir,
 			  gboolean       update,
 			  const char    *password,
 			  FRCompression  compression,
@@ -1039,6 +1155,7 @@ fr_archive_add_directory (FRArchive     *archive,
 	ad_data->archive = archive;
 	ad_data->directory = g_strdup (directory);
 	ad_data->base_dir = g_strdup (base_dir);
+	ad_data->dest_dir = g_strdup (dest_dir);
 	ad_data->update = update;
 	ad_data->password = g_strdup (password);
 	ad_data->compression = compression;
@@ -1065,14 +1182,14 @@ _archive_remove (FRArchive *archive,
 	GList    *scan;
 
 	/* file_list == NULL means delete all files in archive. */
-	
+
 	if (file_list == NULL) {
 		for (scan = archive->command->file_list; scan != NULL; scan = scan->next) {
 			FileData *fdata = (FileData*) scan->data;
 			file_list = g_list_prepend (file_list, fdata->original_path);
 		}
 		file_list_created = TRUE;
-	} 
+	}
 
 	e_file_list = escape_file_list (file_list);
 	fr_command_set_n_files (archive->command, g_list_length (e_file_list));
@@ -1117,11 +1234,9 @@ fr_archive_remove (FRArchive     *archive,
 
 	fr_archive_stoppable (archive, FALSE);
 
-	fr_process_clear (archive->process);
 	fr_command_uncompress (archive->command);
 	_archive_remove (archive, file_list);
 	fr_command_recompress (archive->command, compression);
-	fr_process_start (archive->process);
 }
 
 
@@ -1284,8 +1399,6 @@ fr_archive_extract (FRArchive  *archive,
 		GList *e_file_list;
 
 		e_file_list = escape_file_list (file_list);
-
-		fr_process_clear (archive->process);
 		extract_in_chunks (archive->command,
 				   e_file_list,
 				   dest_dir,
@@ -1293,17 +1406,13 @@ fr_archive_extract (FRArchive  *archive,
 				   skip_older,
 				   junk_paths,
 				   password);
-
 		path_list_free (e_file_list);
 
-		fr_process_start (archive->process);
 		return;
 	}
 
 	/* .. else we have to implement the unsupported options. */
 
-	fr_process_clear (archive->process);
-
 	move_to_dest_dir = (junk_paths 
 			    && ! archive->command->propExtractCanJunkPaths);
 
@@ -1382,10 +1491,7 @@ fr_archive_extract (FRArchive  *archive,
 		char *temp_dir;
 		char *e_temp_dir;
 
-		temp_dir = g_strdup_printf ("%s%s%d",
-					    g_get_tmp_dir (),
-					    "/file-roller.",
-					    getpid ());
+		temp_dir = get_temp_work_dir_name ();
 		ensure_dir_exists (temp_dir, 0700);
 		extract_in_chunks (archive->command,
 				   e_filtered,
@@ -1394,7 +1500,6 @@ fr_archive_extract (FRArchive  *archive,
 				   skip_older,
 				   junk_paths,
 				   password);
-
 		move_files_in_chunks (archive, 
 				      e_filtered, 
 				      temp_dir, 
@@ -1426,8 +1531,6 @@ fr_archive_extract (FRArchive  *archive,
 	if (extract_all) 
 		/* the list has been created in this function. */
 		path_list_free (file_list);
-
-	fr_process_start (archive->process);
 }
 
 
@@ -1454,6 +1557,7 @@ fr_archive_utils__get_file_name_ext (const char *filename)
 {
 	static char * ext[] = {
 		".arj",
+		".bin",
 		".bz", 
 		".bz2", 
 		".ear",
@@ -1462,6 +1566,8 @@ fr_archive_utils__get_file_name_ext (const char *filename)
 		".lzh",
 		".lzo",
 		".rar",
+		".rpm",
+		".sit",
 		".tar", 
 		".tar.bz", 
 		".tar.bz2", 
