@@ -94,8 +94,7 @@ fr_command_cfile_list (FRCommand *comm)
 		fr_process_set_out_line_func (FR_COMMAND (comm)->process, 
 					      list__process_line,
 					      comm);
-		
-		fr_process_clear (comm->process);
+
 		fr_process_begin_command (comm->process, "gzip");
 		fr_process_add_arg (comm->process, "-l");
 		fr_process_add_arg (comm->process, "-q");
@@ -163,18 +162,6 @@ fr_command_cfile_delete (FRCommand *comm,
 }
 
 
-static gchar *
-get_temp_dir_name ()
-{
-	static int count = 0;
-	return g_strdup_printf ("%s%s.%d.%d",
-				g_get_tmp_dir (),
-				"/file-roller-cfile",
-				getpid (),
-				count++);
-}
-
-
 static void
 fr_command_cfile_extract (FRCommand  *comm,
 			  GList      *file_list,
@@ -191,7 +178,7 @@ fr_command_cfile_extract (FRCommand  *comm,
 	char           *temp_file;
 	char           *uncompr_file;
 
-	temp_dir = get_temp_dir_name ();
+	temp_dir = get_temp_work_dir_name ();
 	e_temp_dir = shell_escape (temp_dir);
 
 	/* create a temp dir. */
