@@ -84,63 +84,14 @@ mktime_from_string (char *date_s,
 }
 
 
-static char *
-eat_spaces (char *line)
-{
-	while ((*line == ' ') && (*line != 0))
-		line++;
-	return line;
-}
-
-
-static char **
-split_line (char *line, 
-	    int   n_fields)
-{
-	char **fields;
-	char  *scan, *field_end;
-	int    i;
-
-	fields = g_new0 (char *, n_fields + 1);
-	fields[n_fields] = NULL;
-
-	scan = eat_spaces (line);
-	for (i = 0; i < n_fields; i++) {
-		field_end = strchr (scan, ' ');
-		fields[i] = g_strndup (scan, field_end - scan);
-		scan = eat_spaces (field_end);
-	}
-
-	return fields;
-}
-
-
-static char *
-get_last_field (char *line)
-{
-	int   i;
-	char *field;
-	int   n = 6;
-
-	n--;
-	field = eat_spaces (line);
-	for (i = 0; i < n; i++) {
-		field = strchr (field, ' ');
-		field = eat_spaces (field);
-	}
-
-	return field;
-}
-
-
 static void
 process_line (char     *line, 
 	      gpointer  data)
 {
-	FileData   *fdata;
-	FRCommand  *comm = FR_COMMAND (data);
-	char      **fields;
-	char       *name_field;
+	FileData    *fdata;
+	FRCommand   *comm = FR_COMMAND (data);
+	char       **fields;
+	const char  *name_field;
 
 	g_return_if_fail (line != NULL);
 
@@ -153,7 +104,7 @@ process_line (char     *line,
 
 	/* Full path */
 
-	name_field = get_last_field (line);
+	name_field = get_last_field (line, 6);
 	fields = g_strsplit (name_field, " -> ", 2);
 
 	if (fields[1] == NULL) {
@@ -225,7 +176,6 @@ fr_command_tar_list (FRCommand *comm)
 				      process_line,
 				      comm);
 
-	fr_process_clear (comm->process);
 	fr_process_begin_command (comm->process, "tar");
 	fr_process_add_arg (comm->process, "--force-local");
 	fr_process_add_arg (comm->process, "-tvf");
@@ -241,10 +191,8 @@ process_line__generic (char     *line,
 		       gpointer  data,
 		       char     *action_msg)
 {
-	FRCommand    *comm = FR_COMMAND (data);
-	char         *msg;
-	char         *filename;
-	double        fraction;
+	FRCommand  *comm = FR_COMMAND (data);
+	char       *msg;
 
 	if (line == NULL)
 		return;
@@ -252,15 +200,12 @@ process_line__generic (char     *line,
 	if (line[strlen (line) - 1] == '/') /* ignore directories */
 		return;
 
-	filename = _g_strdup_with_max_size (line, FILENAME_MAX_LENGTH);
-	msg = g_strconcat (action_msg, filename, NULL);
-	g_free (filename);
-
+	msg = g_strconcat (action_msg, line, NULL);
 	fr_command_message (comm, msg);
 	g_free (msg);
 
 	if (comm->n_files != 0) {
-		fraction = (double) comm->n_file++ / comm->n_files;
+		double fraction = (double) comm->n_file++ / comm->n_files;
 		fr_command_progress (comm, fraction);
 	}
 }
@@ -515,9 +460,9 @@ fr_command_tar_recompress (FRCommand     *comm,
 		break;
 	}
 
-	/* Restore original name. */
-
 	if (c_tar->name_modified) {
+		/* Restore original name. */
+		
 		fr_process_begin_command (comm->process, "mv");
 		fr_process_set_sticky (comm->process, TRUE);
 		fr_process_add_arg (comm->process, "-f");
@@ -541,21 +486,137 @@ begin_func__uncompress (gpointer data)
 }
 
 
+static char *
+get_uncompressed_name (FRCommandTar *c_tar,
+		       const char   *e_filename) 
+{
+	char *new_name = g_strdup (e_filename);
+	int   l = strlen (new_name);
+ 
+	switch (c_tar->compress_prog) {
+	case FR_COMPRESS_PROGRAM_NONE:
+		break;
+
+	case FR_COMPRESS_PROGRAM_GZIP:
+		/* X.tgz     -->  X.tar 
+		 * X.tar.gz  -->  X.tar */
+		if (file_extension_is (e_filename, ".tgz")) {
+			new_name[l - 2] = 'a';
+			new_name[l - 1] = 'r';
+		} else if (file_extension_is (e_filename, ".tar.gz")) 
+			new_name[l - 3] = 0;
+		break;
+
+	case FR_COMPRESS_PROGRAM_BZIP:
+		/* X.tbz     -->  X.tar 
+		 * X.tar.bz  -->  X.tar */
+		if (file_extension_is (e_filename, ".tbz")) {
+			new_name[l - 2] = 'a';
+			new_name[l - 1] = 'r';
+		} else if (file_extension_is (e_filename, ".tar.bz")) 
+			new_name[l - 3] = 0;
+		break;
+
+	case FR_COMPRESS_PROGRAM_BZIP2:
+		/* X.tbz2    -->  X.tar 
+		 * X.tar.bz2 -->  X.tar */
+		if (file_extension_is (e_filename, ".tbz2")) {
+			new_name[l - 3] = 'a';
+			new_name[l - 2] = 'r';
+			new_name[l - 1] = 0;
+		} else if (file_extension_is (e_filename, ".tar.bz2")) 
+			new_name[l - 4] = 0;
+		break;
+
+	case FR_COMPRESS_PROGRAM_COMPRESS: 
+		/* X.taz   -->  X.tar 
+		 * X.tar.Z -->  X.tar */
+		if (file_extension_is (e_filename, ".taz")) 
+			new_name[l - 1] = 'r';
+		else if (file_extension_is (e_filename, ".tar.Z")) 
+			new_name[l - 2] = 0;
+		break;
+
+	case FR_COMPRESS_PROGRAM_LZOP:
+		/* X.tzo     -->  X.tar 
+		 * X.tar.lzo -->  X.tar */
+		if (file_extension_is (e_filename, ".tzo")) {
+			new_name[l - 2] = 'a';
+			new_name[l - 1] = 'r';
+		} else if (file_extension_is (e_filename, ".tar.lzo")) 
+			new_name[l - 4] = 0;
+		break;
+	}
+	
+	return new_name;
+}
+
+
+#define MAX_TRIES 50
+
+
+static char *
+get_temp_name (FRCommandTar *c_tar,
+	       const char   *filepath) 
+{
+	char       *dirname = remove_level_from_path (filepath);
+	const char *filename = file_name_from_path (filepath);
+	char       *temp_name = NULL;
+	char       *e_temp_name = NULL;
+	char       *uncomp_temp_name = NULL;
+	static int  count = 0;
+	int         try = 0;
+
+	do {
+		char *tmp_file_name;
+		g_free (temp_name);
+		g_free (uncomp_temp_name);
+		tmp_file_name = g_strdup_printf ("fr.%d.%d.%s", 
+						 getpid (), 
+						 count++, 
+						 filename);
+		temp_name = g_build_filename (dirname, tmp_file_name, NULL);
+		g_free (tmp_file_name);
+		uncomp_temp_name = get_uncompressed_name (c_tar, temp_name);
+	} while ((path_is_file (temp_name)
+		  || path_is_file (uncomp_temp_name))
+		 && (try++ < MAX_TRIES));
+
+	g_free (uncomp_temp_name);
+	g_free (dirname);
+
+	e_temp_name = shell_escape (temp_name);
+	g_free (temp_name);
+
+	return e_temp_name;
+}
+
+
 static void
 fr_command_tar_uncompress (FRCommand *comm)
 {
 	FRCommandTar *c_tar = FR_COMMAND_TAR (comm);
-	gchar        *new_name;
-	gint          l;
+	char         *tmp_name;
 
 	if (c_tar->uncomp_filename != NULL) {
 		g_free (c_tar->uncomp_filename);
 		c_tar->uncomp_filename = NULL;
 	}
-	c_tar->name_modified = FALSE;
 
-	new_name = g_strdup (comm->e_filename);
-	l = strlen (new_name);
+	c_tar->name_modified = c_tar->compress_prog != FR_COMPRESS_PROGRAM_NONE;
+	if (c_tar->name_modified) {
+		tmp_name = get_temp_name (c_tar, comm->filename);
+		if (path_is_file (comm->filename)) {
+			g_print ("test %s [OK]\n", comm->filename);
+			fr_process_begin_command (comm->process, "mv");
+			fr_process_add_arg (comm->process, "-f");
+			fr_process_add_arg (comm->process, comm->e_filename);
+			fr_process_add_arg (comm->process, tmp_name);
+			fr_process_end_command (comm->process);
+		} else
+			g_print ("test %s [NO]\n", comm->filename);
+	} else
+		tmp_name = g_strdup (comm->e_filename);
 
 	switch (c_tar->compress_prog) {
 	case FR_COMPRESS_PROGRAM_NONE:
@@ -567,18 +628,9 @@ fr_command_tar_uncompress (FRCommand *comm)
 			fr_process_set_begin_func (comm->process, begin_func__uncompress, comm);
 			fr_process_add_arg (comm->process, "-f");
 			fr_process_add_arg (comm->process, "-d");
-			fr_process_add_arg (comm->process, comm->e_filename);
+			fr_process_add_arg (comm->process, tmp_name);
 			fr_process_end_command (comm->process);
 		}
-
-		/* X.tgz     -->  X.tar 
-		 * X.tar.gz  -->  X.tar */
-		if (file_extension_is (comm->e_filename, ".tgz")) {
-			c_tar->name_modified = TRUE;
-			new_name[l - 2] = 'a';
-			new_name[l - 1] = 'r';
-		} else if (file_extension_is (comm->e_filename, ".tar.gz")) 
-			new_name[l - 3] = 0;
 		break;
 
 	case FR_COMPRESS_PROGRAM_BZIP:
@@ -587,18 +639,9 @@ fr_command_tar_uncompress (FRCommand *comm)
 			fr_process_set_begin_func (comm->process, begin_func__uncompress, comm);
 			fr_process_add_arg (comm->process, "-f");
 			fr_process_add_arg (comm->process, "-d");
-			fr_process_add_arg (comm->process, comm->e_filename);
+			fr_process_add_arg (comm->process, tmp_name);
 			fr_process_end_command (comm->process);
 		}
-
-		/* X.tbz     -->  X.tar 
-		 * X.tar.bz  -->  X.tar */
-		if (file_extension_is (comm->e_filename, ".tbz")) {
-			c_tar->name_modified = TRUE;
-			new_name[l - 2] = 'a';
-			new_name[l - 1] = 'r';
-		} else if (file_extension_is (comm->e_filename, ".tar.bz")) 
-			new_name[l - 3] = 0;
 		break;
 
 	case FR_COMPRESS_PROGRAM_BZIP2:
@@ -607,19 +650,9 @@ fr_command_tar_uncompress (FRCommand *comm)
 			fr_process_set_begin_func (comm->process, begin_func__uncompress, comm);
 			fr_process_add_arg (comm->process, "-f");
 			fr_process_add_arg (comm->process, "-d");
-			fr_process_add_arg (comm->process, comm->e_filename);
+			fr_process_add_arg (comm->process, tmp_name);
 			fr_process_end_command (comm->process);
 		}
-
-		/* X.tbz2    -->  X.tar 
-		 * X.tar.bz2 -->  X.tar */
-		if (file_extension_is (comm->e_filename, ".tbz2")) {
-			c_tar->name_modified = TRUE;
-			new_name[l - 3] = 'a';
-			new_name[l - 2] = 'r';
-			new_name[l - 1] = 0;
-		} else if (file_extension_is (comm->e_filename, ".tar.bz2")) 
-			new_name[l - 4] = 0;
 		break;
 
 	case FR_COMPRESS_PROGRAM_COMPRESS: 
@@ -627,17 +660,9 @@ fr_command_tar_uncompress (FRCommand *comm)
 			fr_process_begin_command (comm->process, "uncompress");
 			fr_process_set_begin_func (comm->process, begin_func__uncompress, comm);
 			fr_process_add_arg (comm->process, "-f");
-			fr_process_add_arg (comm->process, comm->e_filename);
+			fr_process_add_arg (comm->process, tmp_name);
 			fr_process_end_command (comm->process);
 		}
-
-		/* X.taz   -->  X.tar 
-		 * X.tar.Z -->  X.tar */
-		if (file_extension_is (comm->e_filename, ".taz")) {
-			c_tar->name_modified = TRUE;
-			new_name[l - 1] = 'r';
-		} else if (file_extension_is (comm->e_filename, ".tar.Z")) 
-			new_name[l - 2] = 0;
 		break;
 
 	case FR_COMPRESS_PROGRAM_LZOP:
@@ -646,22 +671,14 @@ fr_command_tar_uncompress (FRCommand *comm)
 			fr_process_set_begin_func (comm->process, begin_func__uncompress, comm);
 			fr_process_add_arg (comm->process, "-dfU");
 			fr_process_add_arg (comm->process, "--no-stdin");
-			fr_process_add_arg (comm->process, comm->e_filename);
+			fr_process_add_arg (comm->process, tmp_name);
 			fr_process_end_command (comm->process);
 		}
-
-		/* X.tzo     -->  X.tar 
-		 * X.tar.lzo -->  X.tar */
-		if (file_extension_is (comm->e_filename, ".tzo")) {
-			c_tar->name_modified = TRUE;
-			new_name[l - 2] = 'a';
-			new_name[l - 1] = 'r';
-		} else if (file_extension_is (comm->e_filename, ".tar.lzo")) 
-			new_name[l - 4] = 0;
 		break;
 	}
 
-	c_tar->uncomp_filename = new_name;
+	c_tar->uncomp_filename = get_uncompressed_name (c_tar, tmp_name);
+	g_free (tmp_name);
 }
 
 
@@ -689,6 +706,7 @@ fr_command_tar_class_init (FRCommandTarClass *class)
 static void 
 fr_command_tar_init (FRCommand *comm)
 {
+	comm->propCanModify                = TRUE;
 	comm->propAddCanUpdate             = TRUE;
 	comm->propAddCanReplace            = FALSE; 
 	comm->propExtractCanAvoidOverwrite = FALSE;
